# R script shown on www.mwolters.com/code/airpollution.html

source("helper_functions.R")                      #-Load functions needed later.
load("RGB.RData")                                 #-Loads data frame RGBdata.
dim(RGBdata)
head(RGBdata)


#=== The Expository Example ====================================================
# Here, we will only use G and B as predcitors, and will only use the first 
# 24750 rows of RGBdata (the ones corresponding to the sub-image shown in the
# chapter).

dat <- RGBdata[1:24750,]  #-Data to use for this example.
smoke <- dat$class==1  #-For selecting only smoke pixels.


#--- Logistic classifier with two predictors ----------
fit <- glm(class ~ G + B, family=binomial(link=logit), data=dat)
summary(fit)

# Make the plot of the data points
windows(width=6, height=5)
par(cex=0.75, mar=c(5,4,0,2)+0.1, family="serif")
plot(dat$G[smoke], dat$B[smoke], xlim=c(0,1), ylim=c(0,1), pch=20, cex=0.25,
     col="blue", xlab="Green", ylab="Blue")
points(dat$G[!smoke],dat$B[!smoke], pch=18, cex=0.25, col="red")

# Add the model's 50% probability contour line
ngrid <- 100
g <- b <- seq(0,1,length.out=ngrid)
pp <- function(g,b) predict(fit, newdata=data.frame(G=g,B=b), type="response")
z <- outer(g, b, pp)
levs <- 0.5
contour(g, b, z, levels=levs, add=TRUE, lwd=2)

# Add the image showing the predictions. Functions addoutline() and 
# mask2outline() are for adding outlines to images, and are found in 
# helper_functions.R.  rImage() is also defined there. It's a small wrapper
# to the rasterImage() function of the graphics base package.
msk <- matrix(dat$class, nrow=150)  #-Create "mask" matrix giving true labels.
FV <- matrix(fit$fitted.values, nrow=150)  #-Fitted probabilities as an image.
FV <- addoutline(FV>0.5, mask2outline(msk, 2))
rImage(FV, c(0, 0.7, 0.33, 1))
text(0.165, 0.7, "model prediction\n(white = smoke)", pos=1,  family="serif")


#--- Expanded feature space ----------
fit2 <- glm(class ~ (G + B + I(G^2) + I(B^2))^2 , 
            family=binomial(link="logit"), data=dat)
summary(fit2)

# Make the plot of the data points
windows(width=6, height=5)
par(cex=0.75, mar=c(5,4,0,2)+0.1, family="serif")
plot(dat$G[smoke], dat$B[smoke], xlim=c(0,1), ylim=c(0,1), pch=20, cex=0.25, 
     col="blue", xlab="Green", ylab="Blue")
points(dat$G[!smoke], dat$B[!smoke], pch=18, cex=0.25, col="red")

# Add the model's 50% probability contour line
ngrid <- 150
g <- b <- seq(0,1,length.out=ngrid)
pp2 <- function(g,b) predict(fit2, newdata=data.frame(G=g,B=b), type="response")
z2 <- outer(g, b, pp2)
levs <- 0.5
contour(g, b, z2, levels=levs, add=TRUE, lwd=2, drawlabels=FALSE)

# Add the image showing the predictions
FV2 <- matrix(fit2$fitted.values, nrow=150)  #-Fitted probabilities as an image.
FV2 <- addoutline(FV2>0.5, mask2outline(msk, 2))
rImage(FV2, c(0, 0.7, 0.33, 1))
text(0.165, 0.7, "model prediction\n(white = smoke)", pos=1,  family="serif")


#--- Try principal components ---------
# Leave out the intercept here, just figuring out the PCs for the covariates
X <- model.matrix(class ~ (G + B + I(G^2) + I(B^2))^2 - 1, data=dat)
S <- cov(X)
p <- dim(S)[1]
eig <- eigen(S)
pve <- eig$values/sum(eig$values)  #-Proportion of variance explained
r <- sum(cumsum(pve)<0.99) + 1     #-Choose first r PCs so that at least 99% var expl.
E <- eig$vectors[, 1:r]            #-The first r eigenvectors
rownames(E) <- colnames(X)         #-Helps interpretation of eigenvectors
A <- X%*%E                         #-Reduced data matrix
nms <- character(length = 0)
for (i in 1:r) nms <- c(nms, paste("a", i, sep=""))
colnames(A) <- nms
newdf <- cbind(data.frame("class"=dat$class), as.data.frame(A))

r               #-Equals 3
head(newdf)     #-Look at the new data

#now repeat the fitting.
fit3 <- glm(class ~ ., family=binomial(link="logit"), data=newdf)
summary(fit3)

# Make the plot of the data points
windows(width=6, height=5)
par(cex=0.75, mar=c(5,4,0,2)+0.1, family="serif")
plot(dat$G[smoke], dat$B[smoke], xlim=c(0,1), ylim=c(0,1), pch=20, cex=0.25, 
     col="blue", xlab="Green", ylab="Blue")
points(dat$G[!smoke], dat$B[!smoke], pch=18, cex=0.25, col="red")

# Add the model's 50% probability contour line
# To do it, we need to make pp3 function take g, b and give the data frame 
# based on PCs.
ngrid <- 150
g <- b <- seq(0,1,length.out=ngrid)
makePC <- function(g,b){
  Xnew <- model.matrix(~ (G + B + I(G^2) + I(B^2))^2 - 1, 
                       data=data.frame(G=g,B=b))
  Anew <- Xnew%*%E
  nms <- character(length = 0)
  for (i in 1:r) nms <- c(nms, paste("a", i, sep=""))
  colnames(Anew) <- nms
  Anew
}
pp3 <- function(g,b) predict(fit3, newdata=as.data.frame(makePC(g,b)), type="response")
z3 <- outer(g, b, pp3)
levs <- 0.5
contour(g, b, z3, levels=levs, add=TRUE, lwd=2, drawlabels=FALSE)

# Add the image showing the predictions
FV3 <- matrix(fit3$fitted.values, nrow=150)
FV3 <- addoutline(FV3>0.5, mask2outline(msk, 2))
rImage(FV3, c(0, 0.7, 0.33, 1))
text(0.165, 0.7, "model prediction\n(white = smoke)", pos=1,  family="serif")


#=== An Extended Example =======================================================
# Now we will use all of the available pixels and do a full analysis including 
# partitioning the data into training, validation, and test sets; as well as 
# model selection to choose a best model.
# 
# For this example, we will use all three color channels R, G, B as well as
# their squared terms, square root terms, and all two-way interactions among 
# powers of different variables. That gives a total of 36 variables to choose
# from. If we seek the best 6-variable model from this set of predictors, we 
# have 1947792 candidate models, which is at least somewhat large.

dat <- RGBdata            #-Data to use for this example.
smoke <- dat$class==1     #-For selecting only smoke pixels.
set.seed(1234)            #-For reproducibility of sampling.

#--- Divide the data -----------
# Assign pixels randomly into training/validation/test groups with a 50/25/25 
# split.  For simplicity just do this at random (don't fix the proportion of 
# smoke/nonsmoke pixels in each group).
npix <- length(smoke)    #-55375 pixels.
pick <- sample(1:npix)   #-Assign random numbers 1 to npix to rows of dat.
cutoff1 <- round(0.5*npix)  #-Rows with numbers 1 to cutoff1 go to training.
cutoff2 <- round(0.75*npix) #-Rows with numbers cutoff1+1 to cutoff2 go to val.
traindf <- dat[pick<=cutoff1, ]
valdf <- dat[(pick>cutoff1)&(pick<=cutoff2), ]
testdf <- dat[pick>cutoff2, ]

# This method of making the model matrix is clunky but does the job. Write it 
# as a function and apply it to the three data frames to get model matrices.
getMM <- function(df){
  model.matrix(
    class ~ (R + G + B + I(R^2) + I(G^2) + I(B^2) 
             + I(R^(1/2)) + I(G^(1/2)) + I(B^(1/2)))^2
             - R:I(R^2) - R:I(R^(1/2)) - I(R^2):I(R^(1/2))
             - G:I(G^2) - G:I(G^(1/2)) - I(G^2):I(G^(1/2))
             - B:I(B^2) - B:I(B^(1/2)) - I(B^2):I(B^(1/2))
             , data=df)
}
Xtrain <- getMM(traindf)
Xval <- getMM(valdf)
Xtest <- getMM(testdf)
Ytrain <- traindf$class  #-true classes for training set.
Yval <- valdf$class      #-true classes for validation set.
Ytest <- testdf$class    #-true classes for test set.

#--- Variable selection method 1: GA ----------
# Use a genetic algorithm (package kofnGA) to search for the best 6-variable
# model.  To do this we need an objective function.  This objective function
# takes a length-six index vector (the columns of X to use in the model) as 
# its first argument, and returns a measure of model quality (validation-set
# deviance) as its result.

# Create the objective function. 
ObjFun <- function(v, X1, Y1, X2, Y2){ 
  # v is the vector of columns to use.
  # X1 is the model matrix for the training data (all columns, and intercept)
  # y1 is the zero/one response vector for the training data
  # X2 is the model matrix for the validation data (all columns, and intercept)
  # y2 is the zero/one response vector for the validation data
  
  M1 <- X1[, c(1,v+1)]    #-Model matrix for the model corresponding to v (train).
  M2 <- X2[, c(1,v+1)]    #-Model matrix for the model corresp. to v (validation).
  
  # Fit the model on the training data. Keep only the coefficients and 
  # convergence flag. Note, we use speedglm.wfit (speeglm package) rather than
  # glm() or glm.fit() because speeglm is considerably faster, even for this 
  # "small" problem size. Note: setting tol.solve=0 is a hack to let the search
  # continue in case of computationally singular matrices.
  fit <- speedglm.wfit(
               y=Y1, X=M1, family=binomial(link="logit"), sparse=FALSE,
               eigendec=FALSE, tol.solve=0)[c("coefficients","convergence")]
  if(!fit$convergence) return(1e20) #-Return bad value if not converged.    
  
  # Use the fitted coefficients to get predicted probabilities for the  
  # validation data. Then compute the deviance (minus 2 times the log-
  # likelihood).
  eta <- M2 %*% fit$coefficients    #-Linear predictor values.
  pp <- exp(eta)/(1+exp(eta))       #-Predicted probabilities.
  pp[pp==1] <- 1 - .Machine$double.eps  #-To avoid -Inf*0 in dev calculation.
  dev <- -2*sum(log(pp)*Y2 + log(1-pp)*(1-Y2))
  
  return(dev)    
}

# Now do the model search using kofnGA (see that package's documentation for 
# more information). 
# === WARNING: === 
#   this step could take time because of repeated model fitting.  Run time is 
#   roughly proportional to ngen*popsize.  As an estimate you could budget 
#   ngen*popsize/500 minutes for this problem.
# ================
require(kofnGA)
require(speedglm)
GAsearch <- kofnGA(n=36, k=6, OF=ObjFun, ngen=250, popsize=50,
                   X1=Xtrain, Y1=Ytrain, X2=Xval, Y2=Yval)
summary(GAsearch)
colnames(Xtrain)[GAsearch$bestsol]    #-Show the variables chosen.
windows(width=6, height=5)
par(cex=0.75, mar=c(5,4,0,2)+0.1, family="serif")
plot(GAsearch)

# Finally, take the best model found and inspect the confusion matrix for 
# prediction on the test set. This part uses function conf(), found in 
# helper_function.R, to create the confusion matrix.  The confusion matrix is 
# given S3 class ConfusionMatrix, with a summary method.
v <- c(1, GAsearch$bestsol + 1)  
M1 <- Xtrain[, v]        #-Model matrix for the best model, using training data.
M3 <- Xtest[, v]         #-Model matrix for the best model, using test data.
fit <- speedglm.wfit(y=Ytrain, X=M1, family=binomial(link="logit"), sparse=FALSE,
                     eigendec=FALSE)
eta <- M3 %*% fit$coefficients    #-Linear predictor values.
pp <- exp(eta)/(1+exp(eta))       #-Predicted probabilities.
prd <- pp>0.5                     #-Test-pixel predictions.
CM <- conf(prd,Ytest)
summary(CM)


#--- Variable selection method 2: LASSO ----------
# First use the glmnet package to get the family of LASSO-regularized models.
# A sequence of models is fit with different values of the regularization 
# parameter lambda. 
require(glmnet)    
fit2 <- glmnet(Xtrain, Ytrain, family="binomial")

# Now, choose the value of lambda that minimizes validation-set deviance. The 
# fitted coefficients with this lambda value define the best model found by this
# method
valdev <- function(p) -2*sum(log(p)*Yval + log(1-p)*(1-Yval))  #-val set deviance
lams <- fit2$lambda                        #-The sequence of lambda values.
deviances <- numeric(length=length(lams))
for(i in 1:length(lams)){
  p <- predict(fit2, newx=Xval, s=lams[i], type="response")
  deviances[i] <- valdev(p)
}
bestloc <- which(deviances==min(deviances, na.rm=TRUE))
bestlam <- lams[bestloc]
bestbeta <- fit2$beta[ , bestloc]   #-Returns all coeffs, even zeroed ones...
bestbeta[bestbeta!=0]   #-Display the selected variables and their coefficients.

# And get the predictions on the test pixels. Show confusion matrix.
pp2 <- predict(fit2, newx=Xtest, s=bestlam, type="response")
prd2 <- pp2>0.5
CM2 <- conf(prd2,Ytest)
summary(CM2)




