# Utility functions used in example_script.R

##########################################################################################
addoutline <- function(img, outline, color=c(1,0,0)){
  # Function to take a greyscale or rgb image (as a matrix or array) plus an outline (as 
  # a logical matrix of matching size), and overlay the outline on the image in color. The 
  # default color is red, but the optional "color" argument can be used to change this.
  
  dims <- dim(img)
  n1 <- dims[1]
  n2 <- dims[2]
  if(length(dims)==2){
    out <- array(0,dim=c(n1,n2,3))
    out[ , , 1] <- img
    out[ , , 2] <- img
    out[ , , 3] <- img
  }else{
    out <- img
  }
  
  out[ , , 1][outline] <- color[1]
  out[ , , 2][outline] <- color[2]
  out[ , , 3][outline] <- color[3]
  
  out
}
##########################################################################################


##########################################################################################
mask2outline <- function(msk, thickness=1, M=NULL){
  # Function to take a logical or 0/1 matrix giving the true classes, and return a logical
  # matrix of the same dimensions, such that the TRUE values of the matrix trace an outline
  # of the class 1 (smoke) regions.  

  n1 <- dim(msk)[1]
  n2 <- dim(msk)[2]
  
  #---Get the 1-pixel thick outline---
  if(thickness>=1){
    tmp1 <- matrix(FALSE,nrow=n1,ncol=n2)
    tmp2 <- matrix(FALSE,nrow=n1,ncol=n2)
    tmp1[2:n1,1:n2] <- msk[2:n1,1:n2] != msk[1:(n1-1),1:n2]  #-North/South
    tmp2[1:n1,2:n2] <- msk[1:n1,2:n2] != msk[1:n1,1:(n2-1)]  #-East/West
    tmp <- tmp1 | tmp2
  }
  
  #---Make the outline thicker, if desired---
  if(thickness>1){
    for(i in 2:thickness){
      tmp1[2:n1,1:n2] <- tmp[2:n1,1:n2] != tmp[1:(n1-1),1:n2]  #-North/South
      tmp2[1:n1,2:n2] <- tmp[1:n1,2:n2] != tmp[1:n1,1:(n2-1)]  #-East/West
      tmp <- tmp | tmp1 | tmp2
    }
  }  
  
  tmp
}
##########################################################################################


##########################################################################################
rImage <- function(img, coord=c(0,0,1,1), ...) {
  # Short version of the rasterImage function to use 0,0,1,1 and interp=F
  rasterImage(img,coord[1],coord[2],coord[3],coord[4],interpolate=FALSE, ...)
}
##########################################################################################


##########################################################################################
conf <- function(p,t){
  # Function to create a confusion matrix from a vector of predictions (p) and a vector of true
  # values (t).  Both p and t are binary or logical.
  out <- matrix(0,nrow=3,ncol=3)
  dimnames(out) <- list(true=c("0","1","total"), predicted=c("0","1","total"))
  out[3,1] <- sum(p==0)
  out[3,2] <- sum(p==1)
  out[1,3] <- sum(t==0)
  out[2,3] <- sum(t==1)
  out[3,3] <- length(t)
  out[1,1] <- sum((t==0) & (p==0))
  out[2,2] <- sum((t==1) & (p==1))
  out[2,1] <- out[3,1] - out[1,1]
  out[1,2] <- out[3,2] - out[2,2]
  out <- out/out[3,3]
  class(out) <- "ConfusionMatrix"
  out
}
##########################################################################################


##########################################################################################
summary.ConfusionMatrix <- function(x) {
  # Summary method for S3 class "ConfusionMatrix"
  out <- list()
  out$matrix <- x
  out$TotErr <- x[1,2] + x[2,1]
  ErrRates <- c(x[1,2]/x[1,3], 
                x[2,1]/x[2,3], 
                (x[1,2]/x[1,3] + x[2,1]/x[2,3])/2 )
  names(ErrRates) <- c("class 0", "class 1", "average")
  out$ErrRates <- ErrRates
  class(out) <- "summary.ConfusionMatrix"
  return(out)
}
##########################################################################################


##########################################################################################
print.summary.ConfusionMatrix <- function(x) {
  # Print method for the summary of a ConfusionMatrix
  cat("\nConfusion matrix:\n")
  print(x$matrix[])
  cat("\nOverall error rate:\n")
  print(x$TotErr)
  cat("\nClasswise error rates:\n")
  print(x$ErrRates)
}
##########################################################################################


